"""
Module provides analysis of robots positions and each frame, calculating various
three-dimensional statistical measures
"""

from multiprocessing.pool import Pool
import os

import numpy as np

RAD2DEG = 180 / np.pi
DEG2RAD = np.pi / 180


def _position_correlation_frame(data_frame: tuple): # pragma: no cover
    kinematics_frame, x_size, y_size = data_frame

    matrix = [[0 for x in range(x_size)] for y in range(y_size)]

    N = len(kinematics_frame)
    distance_matrix = [[(0, 0) for i in range(N)] for j in range(N)]
    for i_ref_bot in range(N):
        sina = np.sin(kinematics_frame[i_ref_bot][1] * DEG2RAD)
        cosa = np.cos(kinematics_frame[i_ref_bot][1] * DEG2RAD)
        for i_bot in range(N):
            if i_bot == i_ref_bot:
                continue
            dx = kinematics_frame[i_bot][2][0] - kinematics_frame[i_ref_bot][2][0]
            dy = kinematics_frame[i_bot][2][1] - kinematics_frame[i_ref_bot][2][1]
            X = round(dy * cosa - dx * sina)
            Y = round(+ dy * sina + dx * cosa)
            distance_matrix[i_ref_bot][i_bot] = (X, Y)

    for i_ref_bot in range(N):
        for i_bot in range(N):
            x, y = distance_matrix[i_ref_bot][i_bot]
            if (0 < x < x_size) and (0 < y < y_size):
                matrix[y][x] += 1 / N**2

    return matrix


def position_correlation(kinematics: list,
                         x_size: int,
                         y_size: int,
                         ) -> list:
    """
    Returns position correlation matrix

    :param kinematics: system's kinematics
    :param x_size: window size for X-axis
    :param y_size: window size for Y-axis
    :return: matrix of scalar values per frame
    """

    data = [(kinematics_frame, x_size, y_size) for kinematics_frame in kinematics]

    with Pool(max(os.cpu_count() - 1, 1)) as pool:
        pc_matrices = pool.map(_position_correlation_frame, data)

    return pc_matrices


def _orientation_correlation_frame(data_frame: tuple): # pragma: no cover
    kinematics_frame, x_size, y_size = data_frame

    matrix = [[0 for x in range(x_size)] for y in range(y_size)]

    N = len(kinematics_frame)
    distance_matrix = [[(0, 0) for i in range(N)] for j in range(N)]
    for i_ref_bot in range(N):
        sina = np.sin(kinematics_frame[i_ref_bot][1] * DEG2RAD)
        cosa = np.cos(kinematics_frame[i_ref_bot][1] * DEG2RAD)
        for i_bot in range(N):
            if i_bot == i_ref_bot:
                continue
            dx = kinematics_frame[i_bot][2][0] - kinematics_frame[i_ref_bot][2][0]
            dy = kinematics_frame[i_bot][2][1] - kinematics_frame[i_ref_bot][2][1]
            X = round(dy * cosa - dx * sina)
            Y = round(+ dy * sina + dx * cosa)
            distance_matrix[i_ref_bot][i_bot] = (X, Y)

    for i_ref_bot in range(N):
        for i_bot in range(N):
            x, y = distance_matrix[i_ref_bot][i_bot]
            if (0 < x < x_size) and (0 < y < y_size):
                matrix[y][x] += np.cos((kinematics_frame[i_bot][1] - kinematics_frame[i_ref_bot][1])
                                       * DEG2RAD) / N**2

    return matrix


def orientation_correlation(kinematics: list,
                            x_size: int,
                            y_size: int,
                            ) -> list:
    """
    Returns orientation correlation matrix

    :param kinematics: system's kinematics
    :param x_size: window size for X-axis
    :param y_size: window size for Y-axis
    :return: matrix of scalar values per frame
    """

    data = [(kinematics_frame, x_size, y_size) for kinematics_frame in kinematics]
    with Pool(max(os.cpu_count() - 1, 1)) as pool:
        oc_matrices = pool.map(_orientation_correlation_frame, data)

    return oc_matrices


def _velocity_correlation_frame(data_frame: tuple): # pragma: no cover
    kinematics_frame, velocities_frame, x_size, y_size = data_frame

    matrix = [[0 for x in range(x_size)] for y in range(y_size)]

    N = len(kinematics_frame)
    distance_matrix = [[(0, 0) for i in range(N)] for j in range(N)]
    for i_ref_bot in range(N):
        sina = np.sin(kinematics_frame[i_ref_bot][1] * DEG2RAD)
        cosa = np.cos(kinematics_frame[i_ref_bot][1] * DEG2RAD)
        for i_bot in range(N):
            if i_bot == i_ref_bot:
                continue
            dx = kinematics_frame[i_bot][2][0] - kinematics_frame[i_ref_bot][2][0]
            dy = kinematics_frame[i_bot][2][1] - kinematics_frame[i_ref_bot][2][1]
            X = round(dy * cosa - dx * sina)
            Y = round(+ dy * sina + dx * cosa)
            distance_matrix[i_ref_bot][i_bot] = (X, Y)

    factor = sum(v[0]**2 + v[1]**2 for v in velocities_frame) / N
    if factor == 0:
        factor = 1
    for i_ref_bot in range(N):
        v = velocities_frame[i_ref_bot]
        for i_bot in range(N):
            u = velocities_frame[i_bot]
            x, y = distance_matrix[i_ref_bot][i_bot]
            if (0 < x < x_size) and (0 < y < y_size):
                matrix[y][x] += 1 * (v[0]*u[0] + v[1]*u[1]) / factor / N**2

    return matrix


def velocity_correlation(kinematics: list,
                         x_size: int,
                         y_size: int,
                         ) -> list:
    """
    Returns velocity correlation matrix

    :param kinematics: system's kinematics
    :param x_size: window size for X-axis
    :param y_size: window size for Y-axis
    :return: matrix of scalar values per frame
    """

    N = len(kinematics[0])
    velocities = [[(0, 0) for i in range(N)]]
    for i_frame in range(1, len(kinematics)):
        velocities_frame = []
        for i_bot in range(N):
            velocities_frame.append((kinematics[i_frame][i_bot][2][0] -
                                     kinematics[i_frame - 1][i_bot][2][0],
                                     kinematics[i_frame][i_bot][2][1] -
                                     kinematics[i_frame - 1][i_bot][2][1]))
        velocities.append(velocities_frame)
    data = [(kinematics[i_frame], velocities[i_frame], x_size, y_size)
            for i_frame in range(1, len(kinematics))]
    with Pool(max(os.cpu_count() - 1, 1)) as pool:
        vc_matrices = pool.map(_velocity_correlation_frame, data)

    return vc_matrices



"""
Module provides the processing of experimental video recordings and identifies ArUco markers
placed on the robots' upper surfaces
"""
from copy import deepcopy
from tqdm import tqdm
import pickle

import numpy as np

import cv2
from matplotlib import pyplot as plt


RAD2DEG = 180 / np.pi
DEG2RAD = np.pi / 180

ARUCO_DICT = {
    "DICT_4X4_50": cv2.aruco.DICT_4X4_50,
    "DICT_4X4_100": cv2.aruco.DICT_4X4_100,
    "DICT_4X4_250": cv2.aruco.DICT_4X4_250,
    "DICT_4X4_1000": cv2.aruco.DICT_4X4_1000,
    "DICT_5X5_50": cv2.aruco.DICT_5X5_50,
    "DICT_5X5_100": cv2.aruco.DICT_5X5_100,
    "DICT_5X5_250": cv2.aruco.DICT_5X5_250,
    "DICT_5X5_1000": cv2.aruco.DICT_5X5_1000,
    "DICT_6X6_50": cv2.aruco.DICT_6X6_50,
    "DICT_6X6_100": cv2.aruco.DICT_6X6_100,
    "DICT_6X6_250": cv2.aruco.DICT_6X6_250,
    "DICT_6X6_1000": cv2.aruco.DICT_6X6_1000,
    "DICT_7X7_50": cv2.aruco.DICT_7X7_50,
    "DICT_7X7_100": cv2.aruco.DICT_7X7_100,
    "DICT_7X7_250": cv2.aruco.DICT_7X7_250,
    "DICT_7X7_1000": cv2.aruco.DICT_7X7_1000,
    "DICT_ARUCO_ORIGINAL": cv2.aruco.DICT_ARUCO_ORIGINAL,
    "DICT_APRILTAG_16h5": cv2.aruco.DICT_APRILTAG_16h5,
    "DICT_APRILTAG_25h9": cv2.aruco.DICT_APRILTAG_25h9,
    "DICT_APRILTAG_36h10": cv2.aruco.DICT_APRILTAG_36h10,
    "DICT_APRILTAG_36h11": cv2.aruco.DICT_APRILTAG_36h11,
}


def calc_angle(point_a: tuple, point_b: tuple) -> float: # pragma: no cover
    """
    Returns angle in degrees between OX-axis and (b-a) vector direction

    :param point_a: vector's begin point
    :param point_b: vector's end point
    :return: angle in degrees
    """
    return RAD2DEG * np.arctan2(point_b[1] - point_a[1], point_b[0] - point_a[0])


def calc_distance(point_a: tuple, point_b: tuple) -> float: # pragma: no cover
    """
    Returns Euclidean distance between two points

    :param point_a: first point
    :param point_b: vector end point
    :return: distance between points
    """

    return ((point_a[0] - point_b[0])**2 + (point_a[1] - point_b[1])**2)**0.5


class Processor:
    """
    *processing.Processor* class provides interface for processing of experiment videos
    """
    def __init__(self):
        self._filename = None
        self._cartesian_kinematics = None
        self._polar_kinematics = None
        self._time = None
        self._aruco_dictionary = cv2.aruco.Dictionary_get(cv2.aruco.DICT_7X7_1000)
        self._aruco_parameters = cv2.aruco.DetectorParameters_create()

    def set_filename(self, filename: str) -> None: # pragma: no cover
        """
        Set path to the file you want to process
        :param filename: the path
        """
        self._filename = filename

    def set_aruco_dict(self, dict_name: str): # pragma: no cover
        """
        Set ArUco dictionary you want to use to
        :param dict_name: the name of the dict
        """
        if dict_name in ARUCO_DICT.keys():
            self._aruco_dictionary = cv2.aruco.Dictionary_get(ARUCO_DICT[dict_name])

    def get_time(self) -> int: # pragma: no cover
        """
        Returns the time parameter

        :return: the time parameter of extracted kinematics
        """
        return self._time

    def cartesian_kinematics(self,
                             bots_number: int,
                             begin_frame: int,
                             end_frame: int,
                             get_each: int,
                             ignore_codes: tuple,
                             scale_parameters: tuple,
                             ) -> list:
        """
        Returns cartesian kinematics for particles in video with given processing parameters

        :param bots_number: number of bots in video
        :param begin_frame: frame to begin the processing
        :param end_frame: frame to end the processing
        :param get_each: frames decimation frequency
        :param ignore_codes: markers to ignore while recognition
        :param scale_parameters: pixels absolute scaling parameters
        :return: list of frame-by-frame particles cartesian kinematic
        """

        alpha, beta = scale_parameters
        video_capture = cv2.VideoCapture(self._filename)

        if begin_frame < 1: # pragma: no cover
            start_frame = 1
        else:
            start_frame = begin_frame

        frames_number = int(video_capture.get(cv2.CAP_PROP_FRAME_COUNT))

        if end_frame > frames_number: # pragma: no cover
            finish_frame = frames_number
        else:
            finish_frame = end_frame

        raw_cart_kin = []
        for current_frame in tqdm(range(start_frame, finish_frame + 1, get_each)):
            video_capture.set(cv2.CAP_PROP_POS_FRAMES, current_frame - 1)
            success, frame = video_capture.read()
            if not success: # pragma: no cover
                raw_cart_kin.append([])
                continue
            frame_converted = cv2.convertScaleAbs(frame, alpha=alpha, beta=beta)
            raw_cart_kin_for_frame = self._raw_cartesian_kinematics_from_frame(frame_converted,
                                                                               ignore_codes)
            raw_cart_kin.append(raw_cart_kin_for_frame)

        completed_cart_kin = self._fill_gaps_in_raw_kinematics(bots_number, raw_cart_kin)
        self._cartesian_kinematics = completed_cart_kin
        self._time = len(completed_cart_kin)
        return completed_cart_kin

    @staticmethod
    def polar_kinematics(cartesian_kinematics: list, field_center: tuple) -> list:
        """
        Returns kinematics extended by a polar angle
        (from 0 to 360 degrees clockwise in relation to X-axis)
        and a distance from field center for each particle

        :param cartesian_kinematics: cartesian kinematics of a system
        :param field_center: a center of a polar coordinates
        :return: polar system's kinematics
        """

        polar_kinematics = deepcopy(cartesian_kinematics)
        for i_frame in tqdm(range(len(polar_kinematics))):
            for i_bot in range(len(polar_kinematics[i_frame])):
                polar_kinematics[i_frame][i_bot] = [
                    polar_kinematics[i_frame][i_bot][0],
                    polar_kinematics[i_frame][i_bot][1],
                    polar_kinematics[i_frame][i_bot][2],
                    calc_angle(field_center, polar_kinematics[i_frame][i_bot][2]),
                    calc_distance(field_center, polar_kinematics[i_frame][i_bot][2])
                ]
        return polar_kinematics

    def field_center_manual(self) -> tuple: # pragma: no cover
        """
        Shows the video's first frame and returns clicked point coordinates

        :return: clicked point coordinates
        """
        video_capture = cv2.VideoCapture(self._filename)
        success, frame = video_capture.read()
        while not success:
            success, frame = video_capture.read()
        center = self._center_of_image(frame)
        return center

    @staticmethod
    def _lines_intersection(first_line_points: tuple, second_line_points: tuple)\
            -> tuple: # pragma: no cover
        """
        Return intersection point of two lines defined by two segments
        """
        p_1, q_1 = first_line_points
        p_2, q_2 = second_line_points
        s_1 = (q_1[0] - p_1[0], q_1[1] - p_1[1])
        s_2 = (q_2[0] - p_2[0], q_2[1] - p_2[1])
        delta = s_2[0] * (-s_1[1]) - (-s_1[0]) * s_2[1]
        delta_2 = (p_1[0] - p_2[0]) * (-s_1[1]) - (-s_1[0]) * (p_1[1] - p_2[1])
        t_2 = delta_2 / delta
        return p_2[0] + t_2 * s_2[0], p_2[1] + t_2 * s_2[1]

    def field_center_auto(self,
                          first_line_markers: tuple,
                          second_line_markers: tuple,
                          scale_parameters: tuple) -> tuple:
        """
        Return center of the field calculated as the intersection of two lines which were defined by
        two pairs of markers

        :param first_line_markers: markers IDs to define the first line
        :param second_line_markers: markers IDs to define the second line
        :param scale_parameters: pixels absolute scaling parameters
        :return: field's center
        """
        alpha, beta = scale_parameters
        video_capture = cv2.VideoCapture(self._filename)
        poses = []
        n_frames_to_try = min(100, int(video_capture.get(cv2.CAP_PROP_FRAME_COUNT)))
        for i_frame in range(n_frames_to_try):
            success, frame = video_capture.read()
            while not success: # pragma: no cover
                success, frame = video_capture.read()
            frame_converted = cv2.convertScaleAbs(frame, alpha=alpha, beta=beta)
            (corners, ids, rejected) = cv2.aruco.detectMarkers(
                frame_converted, self._aruco_dictionary, parameters=self._aruco_parameters
            )
            ids = ids.flatten()
            if set(first_line_markers + second_line_markers).issubset(set(ids.tolist())):
                idx_1 = np.where(ids == first_line_markers[0])
                (top_left, top_right, bottom_right, bottom_left)\
                    = corners[idx_1[0][0]].reshape((4, 2))
                center_1 = ((top_left[0] + bottom_right[0]) // 2,
                            (top_left[1] + bottom_right[1]) // 2)
                idx_2 = np.where(ids == first_line_markers[1])
                (top_left, top_right, bottom_right, bottom_left)\
                    = corners[idx_2[0][0]].reshape((4, 2))
                center_2 = ((top_left[0] + bottom_right[0]) // 2,
                            (top_left[1] + bottom_right[1]) // 2)
                idx_3 = np.where(ids == second_line_markers[0])
                (top_left, top_right, bottom_right, bottom_left)\
                    = corners[idx_3[0][0]].reshape((4, 2))
                center_3 = ((top_left[0] + bottom_right[0]) // 2,
                            (top_left[1] + bottom_right[1]) // 2)
                idx_4 = np.where(ids == second_line_markers[1])
                (top_left, top_right, bottom_right, bottom_left)\
                    = corners[idx_4[0][0]].reshape((4, 2))
                center_4 = ((top_left[0] + bottom_right[0]) // 2,
                            (top_left[1] + bottom_right[1]) // 2)
                poses.append((center_1, center_2))
                poses.append((center_3, center_4))
                break
        if not poses: # pragma: no cover
            return None
        center = self._lines_intersection(poses[0], poses[1])
        return center



    def metric_constant(self, marker_size: float, scale_parameters: tuple) -> float:
        """
        Returns factor that scale distances in pixel on video to distances in centimeters

        :param marker_size: used ArUco marker size in centimeters
        :param scale_parameters: pixels absolute scaling parameters
        :return: scaling factor
        """

        alpha, beta = scale_parameters
        video_capture = cv2.VideoCapture(self._filename)
        success, frame = video_capture.read()
        while not success: # pragma: no cover
            success, frame = video_capture.read()
        frame_converted = cv2.convertScaleAbs(frame, alpha=alpha, beta=beta)
        (corners, ids, rejected) = cv2.aruco.detectMarkers(
            frame_converted, self._aruco_dictionary, parameters=self._aruco_parameters
        )
        if len(corners) == 0: # pragma: no cover
            return None
        marker_corners = corners[0]
        (top_left, top_right, bottom_right, bottom_left) = marker_corners.reshape((4, 2))
        top_right = (int(top_right[0]), int(top_right[1]))
        top_left = (int(top_left[0]), int(top_left[1]))
        metric_constant = marker_size / calc_distance(top_left, top_right)
        return metric_constant

    def _raw_cartesian_kinematics_from_frame(self,
                                             frame: np.ndarray,
                                             ignore_codes: tuple,
                                             ) -> list: # pragma: no cover
        """
        Returns raw cartesian kinematics for particles in frame

        :param frame: frame to process
        :return: raw cartesian kinematics for the given frame
        """
        if frame is None:
            return []
        (corners, ids, rejected) = cv2.aruco.detectMarkers(frame,
                                                           self._aruco_dictionary,
                                                           parameters=self._aruco_parameters
                                                           )
        raw_kinematics_for_frame = []
        recognized_markers_number = len(corners)
        for i in range(recognized_markers_number):
            marker_corners = corners[i]
            marker_id = ids[i][0]
            if marker_id in ignore_codes: # pragma: no cover
                continue
            (top_left, top_right, bottom_right, bottom_left) = marker_corners.reshape((4, 2))
            (top_left, top_right, bottom_right, bottom_left) = (tuple(map(int, top_left)),
                                                                tuple(map(int, top_right)),
                                                                tuple(map(int, bottom_right)),
                                                                tuple(map(int, bottom_left)),
                                                                )
            center_x = (top_left[0] + bottom_right[0]) // 2
            center_y = (top_left[1] + bottom_right[1]) // 2
            top_mid_x = (top_left[0] + top_right[0]) // 2
            top_mid_y = (top_left[1] + top_right[1]) // 2

            angle = calc_angle((center_x, center_y),
                               (top_mid_x, top_mid_y))
            if angle < 0:
                angle = 360 + angle
            # angle increasing corresponds clockwise rotation
            raw_kinematics_for_frame.append([marker_id, angle, (center_x, center_y)])
        return sorted(raw_kinematics_for_frame)

    @staticmethod
    def _center_of_image(image: np.ndarray) -> tuple: # pragma: no cover
        """
        Shows matplotlib window with a given images. Returns last clicked point

        :param image: image to show
        :return: last clicked point
        """
        center = []

        def onclick(event):
            value = (event.xdata, event.ydata)
            center.append(value)
            return center

        fig, ax = plt.subplots()
        plt.imshow(image)
        fig.canvas.mpl_connect("button_press_event", onclick)
        plt.show()
        return center[-1]

    @staticmethod
    def _fill_gaps_in_raw_kinematics(bots_number: int, raw_cartesian_kinematics: list)\
            -> list: # pragma: no cover
        """
        Returns cartesian kinematics with filling gaps from unrecognized bots
        by they future positions

        :param bots_number: total number of particles in video
        :param raw_cartesian_kinematics: raw Cartesian kinematics
        :return: Cartesian kinematics with filled gaps
        """
        raw_kinematics = deepcopy(raw_cartesian_kinematics)

        frames_number = len(raw_kinematics)
        best_recognized_frame_number = 0

        for i in range(1, frames_number):
            if len(raw_kinematics[i]) > len(raw_kinematics[best_recognized_frame_number]):
                best_recognized_frame_number = i

        top_recognized_bots_number = len(raw_kinematics[best_recognized_frame_number])

        assert  top_recognized_bots_number <= bots_number, 'Number of recognized markers exceeded the expected value. Kinematics processing was aborted!'
        assert  top_recognized_bots_number >= bots_number, 'Number of recognized markers did not reach the expected value. Kinematics processing was aborted!'

        total_ids = np.array(raw_kinematics[best_recognized_frame_number])[:, 0]
        for i_frame in range(frames_number):
            if len(raw_kinematics[i_frame]) != bots_number:
                if len(raw_kinematics[i_frame]) == 0:
                    current_ids = np.array([])
                else:
                    current_ids = np.array(raw_kinematics[i_frame])[:, 0]
                difference = list(set(total_ids) - set(current_ids))
                for i_absent_bot in range(len(difference)):
                    for i_next_frame in range(i_frame + 1, frames_number):
                        bot_searched_out = False
                        if len(raw_kinematics[i_next_frame]) == 0:
                            new_bots_ids = np.array([])
                        else:
                            new_bots_ids = np.array(raw_kinematics[i_next_frame])[:, 0]
                        if difference[i_absent_bot] not in set(new_bots_ids):
                            continue
                        for i_new_bot in range(len(raw_kinematics[i_next_frame])):
                            if raw_kinematics[i_next_frame][i_new_bot][0] ==\
                                    difference[i_absent_bot]:
                                raw_kinematics[i_frame].append(
                                    raw_kinematics[i_next_frame][i_new_bot])
                                bot_searched_out = True
                                break
                        if bot_searched_out:
                            break
        complete_kinematics = [sorted(raw_kinematics[i_frame].copy())
                               for i_frame in range(frames_number)
                               if (len(raw_kinematics[i_frame])) == bots_number]
        return complete_kinematics

    @staticmethod
    def load_p(filename) -> list:
        """
        Load system's kinematics serialized by *pickle*
        :param filename: path to *.pickle* file
        :return: system's kinematics
        """
        with open(filename, 'rb') as file:
            kin = list(pickle.load(file))
            if not kin: # pragma: no cover
                return None
            for i, bot in enumerate(kin):
                if len(bot) != len(kin[0]): # pragma: no cover
                    return None
            return kin



"""
Module provides analysis of robots positions and each frame, calculating various
two-dimensional statistical measures
"""

import multiprocessing as mp
import os

import numpy as np


RAD2DEG = 180 / np.pi
DEG2RAD = np.pi / 180


def _orientation_angles(kinematics: list) -> list: # pragma: no cover
    angles = list(np.array(kinematics, dtype=object)[:, :, 1])
    return angles


def _positions(kinematics: list) -> list:
    positions = list(np.array(kinematics, dtype=object)[:, :, 2])
    return positions


def _polar_angles(kinematics: list) -> list:
    polar_angles = list(np.array(kinematics, dtype=object)[:, :, 3])
    return polar_angles


def _distances_from_center(kinematics: list) -> list:
    distances = list(np.array(kinematics, dtype=object)[:, :, 4])
    return distances


def _sum_points(p1: tuple, p2: tuple, scale: float = 1) -> tuple: # pragma: no cover
    return (p1[0] + p2[0] * scale,
            p1[1] + p2[1] * scale)


def _cross_product(v: tuple, u: tuple) -> float: # pragma: no cover
    return v[0]*u[1] - v[1]*u[0]


def calc_angle(point_a: tuple, point_b: tuple) -> float:
    """
    Returns angle in degrees between OX-axis and (b-a) vector direction

    :param point_a: vector's begin point
    :param point_b: vector's end point
    :return: scalar value
    """

    return RAD2DEG * np.arctan2(point_b[1] - point_a[1], point_b[0] - point_a[0])


def calc_distance(point_a: tuple, point_b: tuple) -> float:
    """
    Returns Euclidean distance between two points

    :param point_a: first point
    :param point_b: vector end point
    :return: scalar value
    """

    return ((point_a[0] - point_b[0])**2 + (point_a[1] - point_b[1])**2)**0.5


def mean_distance_from_center(kinematics: list) -> list:
    """
    Returns average particles' distance from center for each frame

    :param kinematics: system's kinematics extended by polar coordinates
    :return: scalar value for each frame
    """

    mean_distance = np.array(_distances_from_center(kinematics), dtype=float).mean(axis=1).tolist()
    return mean_distance


def mean_polar_angle(kinematics: list) -> list:
    """
    Returns average particles' polar angle for each frame. Angle for certain frame calculating
    by accumulating all angle changes in previous frames

    :param kinematics: system's kinematics extended by polar coordinates
    :return: list of scalar values
    """

    angles = _polar_angles(kinematics)
    mean_angles_raw = np.array(angles, dtype=float).mean(axis=1)
    mean_angles_corrected = [mean_angles_raw[0]]
    for i_frame in range(1, len(mean_angles_raw)):
        difference = mean_angles_raw[i_frame] - mean_angles_raw[i_frame - 1]
        if abs(difference) < 100: # pragma: no cover
            mean_angles_corrected.append(mean_angles_corrected[-1] + difference)
        else:
            if difference > 0: # pragma: no cover
                mean_angles_corrected.append(mean_angles_corrected[-1] + difference - 360)
            else: # pragma: no cover
                mean_angles_corrected.append(mean_angles_corrected[-1] - difference + 360)
    return mean_angles_corrected


def mean_polar_angle_absolute(kinematics: list):
    """
    Returns average particles' polar angle for each frame. Angle for certain frame calculating
    by accumulating absolute values of all angle changes in previous frames

    :param kinematics: system's kinematics extended by polar coordinates
    :return: list of scalar values
    """

    angles = _polar_angles(kinematics)
    mean_angles_raw = np.array(angles, dtype=float).mean(axis=1)
    mean_angles_accumulated = [mean_angles_raw[0]]
    for i_frame in range(1, len(mean_angles_raw)):
        absolute_difference = abs(mean_angles_raw[i_frame] - mean_angles_raw[i_frame - 1])
        if absolute_difference < 100: # pragma: no cover
            mean_angles_accumulated.append(mean_angles_accumulated[-1] + absolute_difference)
        else: # pragma: no cover
            mean_angles_accumulated.append(mean_angles_accumulated[-1] - absolute_difference + 360)
    return mean_angles_accumulated


def mean_cartesian_displacements(kinematics: list):
    """
    Returns average particles' cartesian displacement for each frame

    :param kinematics: system's kinematics
    :return: list of scalar values
    """

    positions = _positions(kinematics)
    first_frame_positions = positions[0]
    mcd = []
    for i_frame in range(len(positions)):
        current_frame_positions = positions[i_frame]
        current_total_cartesian_displacement = 0
        for i_bot in range(len(first_frame_positions)):
            current_total_cartesian_displacement += \
                calc_distance(current_frame_positions[i_bot], first_frame_positions[i_bot])
        current_mcd = \
            current_total_cartesian_displacement / len(first_frame_positions)
        mcd.append(current_mcd)
    return mcd


def _local_bond_orientation(folds_number: int, bot_position: tuple, neighbours_positions: list):
    p = 0
    for neighbour_index in range(len(neighbours_positions)):
        p += np.exp(1j * folds_number *
                    calc_angle(bot_position, neighbours_positions[neighbour_index]) * DEG2RAD)
    return np.absolute(p) / len(neighbours_positions)


def bond_orientation(kinematics: list,
                     neighbours_number: int,
                     folds_number: int,
                     get_each: int = 1) -> list:
    """
    Returns bond orientation order parameter

    :param kinematics: system's kinematics
    :param neighbours_number: number of neighbours to use in calculations
    :param folds_number: number of folds to use in calculations
    :param get_each: frames decimation frequency
    :return: list of scalar values
    """

    boo = []
    positions = _positions(kinematics)
    for i_frame in range(0, len(kinematics), get_each):
        current_frame_boo = 0
        current_frame_positions = positions[i_frame]
        for i_bot in range(len(kinematics[i_frame])):
            reference_bot_position = current_frame_positions[i_bot]
            neighbours_positions = list(current_frame_positions.copy())
            neighbours_positions.sort(key=lambda pos: calc_distance(reference_bot_position, pos))
            neighbours_positions = neighbours_positions[:neighbours_number]
            local_boo = \
                _local_bond_orientation(folds_number, reference_bot_position, neighbours_positions)
            current_frame_boo += local_boo
        current_frame_boo /= len(kinematics[i_frame])
        boo.append(current_frame_boo)
    return boo


def chi_4(kinematics: list,
          tau: int,
          a: float) -> float:
    """
    Returns spatio-temporal correlation parameter chi_4 for given time and space gape

    :param kinematics: system's kinematics
    :param tau: characteristic time in frames
    :param a: characteristic distance in pixels
    :return: scalar value
    """

    q_sequence = []
    N = len(kinematics[0])
    for i_frame in range(len(kinematics) - tau):
        q = 0
        for i_bot in range(N):
            q += int(a - calc_distance(kinematics[i_frame + tau][i_bot][2],
                                       kinematics[i_frame][i_bot][2]) >= 0)
        q /= N
        q_sequence.append(q)
    t_corr = N * np.std(q_sequence)
    return t_corr


def _is_collide(bot_1: list, bot_2: list, d: int = 300) -> bool:  # pragma: no cover
    return calc_distance(bot_1[2], bot_2[2]) <= d


def _adjacency_matrix(kinematics_frame: list, collide_function) -> list: # pragma: no cover
    N = len(kinematics_frame)
    adj_matrix = [[0 for i in range(N)] for j in range(N)]
    for i_bot in range(N):
        for j_bot in range(N):
            if i_bot == j_bot:
                continue
            adj_matrix[i_bot][j_bot] = \
                int(collide_function(kinematics_frame[i_bot], kinematics_frame[j_bot]))
    return adj_matrix


def _clustering_coefficient_frame(data_frame: tuple) -> float: # pragma: no cover
    kinematics_frame, collide_function = data_frame
    N = len(kinematics_frame)
    adj_matrix = _adjacency_matrix(kinematics_frame, collide_function)
    cl_coeff = 0
    for i in range(N):
        k_i = sum(adj_matrix[i][j] for j in range(N))
        if k_i in (0, 1):
            continue
        c_i = 0
        for j in range(N):
            for k in range(N):
                c_i += adj_matrix[i][j] * adj_matrix[j][k] * adj_matrix[k][i]
        c_i /= k_i * (k_i - 1)
        cl_coeff += c_i
    cl_coeff /= N
    return cl_coeff


def cluster_dynamics(kinematics: list,
                     collide_function=_is_collide) -> list:
    """
    Returns collision graph average clustering coefficient

    :param kinematics: system's kinematics
    :param collide_function: collision detection function
    :return: list of scalar values
    """

    cl_coeff_seq = []
    data = [(kinematics[i_frame], collide_function) for i_frame in range(len(kinematics))]
    with mp.Pool(max(os.cpu_count() - 1, 1)) as pool:
        cl_coeff_seq = pool.map(_clustering_coefficient_frame, data)
    return cl_coeff_seq



"""
Module provides the dashboard with evolution of system's parameters
"""
import math
from tqdm import tqdm
from copy import deepcopy

import matplotlib.gridspec as gridspec
import numpy as np
import cv2
from cv2 import aruco
from celluloid import Camera
from matplotlib import pyplot as plt


# parameter for the ArUco recognition & processing

size_of_marker = 0.03
mtx = np.array(
    [
        [2.12753367e03, 0.00000000e00, 7.24147082e02],
        [0.00000000e00, 2.12169339e03, 7.65231051e02],
        [0.00000000e00, 0.00000000e00, 1.00000000e00],
    ]
)

dist = np.array(
    [
        [-4.76057214e01],
        [3.72141657e03],
        [-1.18878543e-02],
        [9.99370683e-03],
        [2.23865999e03],
        [-4.89979090e01],
        [3.86516794e03],
        [-6.64583433e02],
        [0.00000000e00],
        [0.00000000e00],
        [0.00000000e00],
        [0.00000000e00],
        [0.00000000e00],
        [0.00000000e00],
    ]
)

ARUCO_DICT = {
    "DICT_4X4_50": cv2.aruco.DICT_4X4_50,
    "DICT_4X4_100": cv2.aruco.DICT_4X4_100,
    "DICT_4X4_250": cv2.aruco.DICT_4X4_250,
    "DICT_4X4_1000": cv2.aruco.DICT_4X4_1000,
    "DICT_5X5_50": cv2.aruco.DICT_5X5_50,
    "DICT_5X5_100": cv2.aruco.DICT_5X5_100,
    "DICT_5X5_250": cv2.aruco.DICT_5X5_250,
    "DICT_5X5_1000": cv2.aruco.DICT_5X5_1000,
    "DICT_6X6_50": cv2.aruco.DICT_6X6_50,
    "DICT_6X6_100": cv2.aruco.DICT_6X6_100,
    "DICT_6X6_250": cv2.aruco.DICT_6X6_250,
    "DICT_6X6_1000": cv2.aruco.DICT_6X6_1000,
    "DICT_7X7_50": cv2.aruco.DICT_7X7_50,
    "DICT_7X7_100": cv2.aruco.DICT_7X7_100,
    "DICT_7X7_250": cv2.aruco.DICT_7X7_250,
    "DICT_7X7_1000": cv2.aruco.DICT_7X7_1000,
    "DICT_ARUCO_ORIGINAL": cv2.aruco.DICT_ARUCO_ORIGINAL,
    "DICT_APRILTAG_16h5": cv2.aruco.DICT_APRILTAG_16h5,
    "DICT_APRILTAG_25h9": cv2.aruco.DICT_APRILTAG_25h9,
    "DICT_APRILTAG_36h10": cv2.aruco.DICT_APRILTAG_36h10,
    "DICT_APRILTAG_36h11": cv2.aruco.DICT_APRILTAG_36h11,
}


def get_video(filename, begin_frame, end_frame, get_each):
    """
        Returns a list with the frames of the input video

        :param filename: the path
        :param bots_number: number of bots in video
        :param begin_frame: frame to begin the processing
        :param end_frame: frame to end the processing
        :param get_each: frames decimation frequency
        :return: list with the frames of the input video
    """

    video_capture = cv2.VideoCapture(filename)

    if begin_frame < 1: # pragma: no cover
        start_frame = 1
    else:
        start_frame = begin_frame

    frames_number = int(video_capture.get(cv2.CAP_PROP_FRAME_COUNT))

    if end_frame > frames_number: # pragma: no cover
        finish_frame = frames_number
    else:
        finish_frame = end_frame

    frames = []
    raw_cart_kin = []
    for current_frame in tqdm(range(start_frame, finish_frame + 1, get_each)):
        video_capture.set(cv2.CAP_PROP_POS_FRAMES, current_frame - 1)
        success, frame = video_capture.read()
        if success: # pragma: no cover
            frames.append(frame)
    return frames

def save_video(output_name: str, frames: list, framerate: int = 50) -> None:
    """
       Saves a list with the frames os a video file
        :param output_name: name of the output file
        :param frames: list of the frames from the get_video method output
        :param framerate: frames per second
    """
    frame_size = frames[0].shape[0:2][::-1]

    fourcc = cv2.VideoWriter_fourcc(*"XVID")
    out = cv2.VideoWriter(output_name, fourcc, framerate, frame_size)
    for frame in frames:
        out.write(frame)
    out.release()

    print(f"Video saved as {output_name}")

def create_dashboard(video: list,
                  output_name:str,
                  cart_disp:list,
                  boo:list,
                  stcp:list, 
                  cl_coeff:list,
                  distance:list,
                  angle:list,
                  angle_abs:list,
                  fps:int,
                 ) -> None:
    """
    Creates .gif with simulteneous evolution of the system parameters along with еру original video

    :param video: list of the frames from the get_video method output
    :param output_name: name of the output file
    :param cart_disp: cartesian displacement
    :param boo: bond orientational order parameter
    :param stcp: spatio-temporal correlation parameter
    :param cl_coeff: clustering coefficient
    :param distance: mean distance from the center
    :param angle: mean polar angle
    :param angle_abs: absolute value of the mean polar angle
    :param fps: frames per second
    """
    
    fig, ax = plt.subplots()
    fig = plt.figure(layout="constrained", figsize = (12,8))
    
    gs = gridspec.GridSpec(4, 4, figure=fig)
    
    ax1 = fig.add_subplot(gs[:3, :3])
    ax2 = fig.add_subplot(gs[:1, -1])
    ax3 = fig.add_subplot(gs[1:2, -1])
    ax4 = fig.add_subplot(gs[2:3, -1])
    ax5 = fig.add_subplot(gs[3:4, -1])
    ax6 = fig.add_subplot(gs[-1, 0])
    ax7 = fig.add_subplot(gs[-1, 1])
    ax8 = fig.add_subplot(gs[-1, 2])

    camera = Camera(fig)
    
    time = [i for i in range(len(cart_disp))]

    for i in tqdm(range(len(video))):

        ax1.imshow(video[i])
        ax1.text(x = 10, y = 30, s=f'Frame {i}, {round(i/fps, 2)} sec', color ='white')

        ax1.set_title('Source video')

        ax2.set_title('Cartesian displacement')

        ax3.set_title('Bond orientation parameter')

        ax4.set_title('S-t correlation parameter')

        ax5.set_title('Clustering coefficient')

        ax6.set_title('Mean distance from the center')

        ax7.set_title('Mean polar angle')

        ax8.set_title('Mean polar angle path')

        axs = [ax2, ax3, ax4, ax5, ax6, ax7, ax8]
        data = [cart_disp, boo, stcp, cl_coeff, distance, angle, angle_abs]

        for d, ax in zip(data, axs):
            ax.plot(time[0:i], d[0:i], color="blue")
            ax.axis(xmin = min(time), xmax = max(time), ymin = min(d), ymax = max(d) + 
                 (max(d) - min(d))/10)

        camera.snap()

    animation = camera.animate()
    animation.save(output_name)

def draw_markers(frames: list,
                 marker_type: str = "DICT_7X7_1000"
                 ) -> list:
    """
        Returns the list with the frames and highlighted markers.

        :param frames: list of the frames from the get_video method output
        :param marker_type: Mmrker type of robots
    """
    frames_altered = deepcopy(frames)
    for j in tqdm(range(len(frames_altered))):
        image = frames_altered[j]
        arucoDict = cv2.aruco.Dictionary_get(ARUCO_DICT[marker_type])
        arucoParams = cv2.aruco.DetectorParameters_create()

        (corners, ids, rejected) = cv2.aruco.detectMarkers(
            image, arucoDict, parameters=arucoParams
        )

        rvecs, tvecs, trash = aruco.estimatePoseSingleMarkers(
            corners, size_of_marker, mtx, dist
        )

        if len(corners) != 0:

            for k in range(len(corners)):
                markerCorner = corners[k]
                markerID = ids[k]
                (topLeft, topRight, bottomRight, bottomLeft) = markerCorner.reshape(
                    (4, 2)
                )

                topRight = (int(topRight[0]), int(topRight[1]))
                bottomRight = (int(bottomRight[0]), int(bottomRight[1]))
                bottomLeft = (int(bottomLeft[0]), int(bottomLeft[1]))
                topLeft = (int(topLeft[0]), int(topLeft[1]))
                contour = np.array(
                    [
                        list(topLeft),
                        list(bottomLeft),
                        list(bottomRight),
                        list(topRight),
                    ],
                    dtype=np.int32,
                )

                triangle = np.array(
                    [
                        [
                            (list(topLeft)[0] + 3 * list(topRight)[0]) / 4,
                            (list(topLeft)[1] + 3 * list(topRight)[1]) / 4,
                        ],
                        [
                            (list(bottomRight)[0] + 3 * list(topRight)[0]) / 4,
                            (list(bottomRight)[1] + 3 * list(topRight)[1]) / 4,
                        ],
                        list(topRight),
                    ],
                    dtype=np.int32,
                )

                cv2.fillPoly(image, pts=[contour], color=(179, 0, 0))
                cv2.fillPoly(image, pts=[triangle], color=(255, 179, 0))

                cX, cY = int((topLeft[0] + bottomRight[0]) / 2.0), int((topLeft[1] + bottomRight[1]) / 2.0)

                angle = math.atan2(topRight[0] - cX, topRight[1] - cY)
                if angle < 0:
                    angle = 2 * np.pi + angle
                angle = np.degrees(angle)

                font = cv2.FONT_HERSHEY_DUPLEX
                textsize = cv2.getTextSize(str(markerID), font, 1, 2)[0]
                textX = int((cX - textsize[0] / 4.5))
                cv2.putText(
                    image, str(markerID), (textX, cY + 5), font, 0.5, (255, 255, 255), 2
                )
    return frames_altered